package com.interdataworking.gateway;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import com.interdataworking.*;
import com.interdataworking.util.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import java.util.*;

/**
 * By default, all statements in the body of the rule are discarded
 *
 * RULE 1:
 *
 * (id, RDF.type, RDF.Statement)
 * (id, RDF.subject, s)
 * (id, RDF.predicate, p)
 * (id, RDF.object, o)
 * =>
 * remove the original four
 * replace every id by (s, p, o)
 *
 */

public class Quad2Reification extends SimpleModelGateway {

  HashMap map;

  public void init(Model srcMD, Model destMD) {
  }

  public Model translate(Model src) throws ModelException {

    // first determine all mappings IDs to statements
    map = new HashMap();
    Model m = src.duplicate();
    this.dest = src;

    for(Enumeration en = src.find(null, RDF.type, RDF.Statement).elements(); en.hasMoreElements();) {

      Resource id = ((Statement)en.nextElement()).subject();
      // get its s, p, o
      Resource s = RDFUtil.getObjectResource(src, id, RDF.subject);
      Resource p = RDFUtil.getObjectResource(src, id, RDF.predicate);
      RDFNode o = RDFUtil.getObjectResource(src, id, RDF.object);

      if(s != null && p != null && o != null) {

	map.put(id, stmt(s, p, o));

	m.remove(stmt(id, RDF.type, RDF.Statement));
	m.remove(stmt(id, RDF.subject, s));
	m.remove(stmt(id, RDF.predicate, p));
	m.remove(stmt(id, RDF.object, o));
      }
    }

    return super.translate(m);
  }

  Resource replace(Resource r) throws ModelException {

    Resource st = (Resource)map.get(r);
    return st != null ? st : r;
  }

  protected void processStatement(Statement st) throws ModelException {

    Resource s = replace(st.subject());
    Resource p = replace(st.predicate());
    RDFNode o = (st.object() instanceof Resource ? replace((Resource)st.object()) : st.object());

    dest.remove(st);
    dest.add(stmt(s, p, o));
  }
  
}
