package com.interdataworking.gateway;

import java.io.*;
import java.util.*;
import org.w3c.rdf.model.*;
import edu.stanford.db.xml.util.*;
import edu.stanford.db.rdf.syntax.generic.*;
import org.xml.sax.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.uml.vocabulary.uml_core_20000422.UMLCore;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;


public class XMI2Obj extends GenericXML2RDF {

  static final QName XMI_Content = createQName(null, "XMI.content");
  static final QName XMI_ID = createQName(null, "xmi.id");
  static final QName XMI_UUID = createQName(null, "xmi.uuid");
  static final QName XMI_IDREF = createQName(null, "xmi.idref");
  static final QName XMI_VALUE = createQName(null, "xmi.value");

  // -1 while skipping
  int depth = -1;

  //  Object subject; // Resource or string (xmi.id)
  //  Resource predicate;

  HashMap idmap = new HashMap(); // maps a string (xml.id) to a resource
  
  HashSet todo = new HashSet(); // contains IncompleteStatements


  class IncompleteStatement {

    Object s; // String (xmi.id) or Resource
    Resource p;
    Object o; // String (xmi.id) or RDFNode
    int order;

    public IncompleteStatement(Object s, Resource p, Object o, int order) {
      this.s = s;
      this.p = p;
      this.o = o;
      this.order = order;
    }
  }

  /** the resource contains predicate for property elements and subject for class/instance elements */
  class ResourceElement extends Element {

    Object resource;
    
    void setResource(Object r) {
      resource = r;
    }
    Object getResource() {
      return resource;
    }
  }

  void checkTodo(String id, Resource r) throws ModelException {

    Iterator it = todo.iterator();
    while(it.hasNext()) {

      IncompleteStatement st = (IncompleteStatement)it.next();
      if(id.equals(st.s))
	st.s = r;
      if(id.equals(st.o))
	st.o = r;
      // could resolve all?
      if(st.s instanceof Resource && st.o instanceof RDFNode) {
	// complete
	it.remove();
	Statement stmt = createStatement((Resource)st.s, st.p, (RDFNode)st.o);
	if(st.order > 0) { // order by reification
	  createStatement(stmt, RDFX.order, nodeFactory.createLiteral(st.order));
	}
      }
    }
  }

  protected Resource createResource(String ns, String ln) throws ModelException {

    if(ln.startsWith("Foundation.")) {
      int p = ln.indexOf('.');
      int i = ln.indexOf('.', p+1);
      return super.createResource("http://www.omg.org/uml/1.3/" + ln.substring(0, i+1),
				  ln.substring(i+1));
    } else
      return super.createResource(ns, ln);
  }


  // returns string (xmi.id) or resource

  Object getInstance(Element el) throws ModelException {

    String uuid = el.getAttribute(XMI_UUID);
    String id = el.getAttribute(XMI_ID);

    if(id != null) {
      Resource r = uuid != null ? createResource(uuid) : nodeFactory.createUniqueResource();
      idmap.put(id, r);
      checkTodo(id, r);
      return r;
    }

    String ref = el.getAttribute(XMI_IDREF);
    Resource r = (Resource)idmap.get(ref);
    return r == null ? (Object)ref : (Object)r;
  }
  
  /** @return null if xmi.idref is defined */
  Resource getClass(Element el) throws ModelException {

    return el.getAttribute(XMI_IDREF) != null ? null : createResource(el.getName());
  }
  
  Resource getProperty(Element el) throws ModelException {

    return createResource(el.getName());
  }
  
  Literal getPropertyValue(Element el) throws ModelException {

    String val = el.getAttribute(XMI_VALUE);
    return val != null ? createLiteral(val) : null;
  }

  void createSafeStatement(Object s, Resource p, Object o, int order) throws ModelException {

    if(s instanceof Resource && o instanceof RDFNode) {
      Statement st = createStatement((Resource)s, p, (RDFNode)o);
      // order by reification
      if(order > 0)
	createStatement(st, RDFX.order, nodeFactory.createLiteral(order));
    } else
      todo.add(new IncompleteStatement(s, p, o, order));
  }

  public Element createElement() {
    return new ResourceElement();
  }

  
  public void startElement (String uri, String name, String qname, Attributes al) throws SAXException {
    
    _startElement(uri, name, al);

    try {

      ResourceElement current = (ResourceElement)this.current;

      if(depth < 0) {
	// watch for XMI.content
	if(XMI_Content.equals(current.getName())) {
	  depth = 0;
	}
      } else {
	if(depth % 2 == 0) {
	  // CLASSNAME
	  Object s = getInstance(current);
	  current.setResource(s);
	  Resource c = getClass(current);
	  Object subject = null;

	  if(depth > 0) {
	    // finish previous statement

	    // determine the ordinal number of this predicate
	    int order = current.getNumInParent();
	    //	    System.out.println("ORDER of " + current.getName() + " within parent is " + order);

	    Resource predicate = (Resource)((ResourceElement)current.getParent()).getResource();
	    subject = ((ResourceElement)((ResourceElement)current.getParent()).getParent()).getResource();

	    createSafeStatement(subject, predicate, s, order);
	  }

	  if(c != null) // subject must be non-null
	    createSafeStatement(s, RDF.type, c, 0);
	  
	} else {
	  // depth % 2 == 1
	  // PROPERTY
	  Resource predicate = getProperty(current);
	  current.setResource(predicate);

	  // is value in place?
	  RDFNode value = getPropertyValue(current);
	  if(value != null) {
	    Object subject = ((ResourceElement)current.getParent()).getResource();
	    createSafeStatement(subject, predicate, value, 0);
	  }
	}
	depth++;
      }

    } catch (Exception exc) { // ModelException or DigestException
      throw new SAXException(exc);
    }
    //    System.out.println("Start element: " + current.getName());
  }

  public void endElement (String uri, String name, String qName) throws SAXException {

    depth--;

    try {

      ResourceElement current = (ResourceElement)this.current;

      if(depth % 2 == 1) {
	// PREDICATE
	// check if this element has CDATA content (only?)

	if((current.getType() & Element.CDATA) != 0) {

	  //	  System.out.println("Type of " + current.getName() + " is " + current.getType());
	  // create statement with a literal value
	  Resource predicate = (Resource)current.getResource();
	  Object subject = ((ResourceElement)current.getParent()).getResource();

	  //	  System.out.println("Sub " + subject + ", pred " + predicate + ", obj " + current.getValue());

	  createSafeStatement(subject, predicate, nodeFactory.createLiteral(current.getValue()), 0);
	}
      }
    } catch (Exception exc) {
      throw new SAXException(exc);
    }

    _endElement(uri, name);
  }
  
  public static void main (String args[]) throws Exception {

    if(args.length != 1) {
      System.err.println("Usage: java [-Dorg.xml.sax.parser=<classname>] XMI2Obj " +
			 //			 "[-fetch_schemas | -f] [-stream | -s] [-robust | -r] [ URI | filename ]");
			 "[ URI | filename ]");
      System.exit(1);
    }

    XMI2Obj gp;
    _main(args[0], gp = new XMI2Obj());
  }
}
