package com.interdataworking.gateway.sqlddl;

import java.io.*;
import java.util.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;

import com.interdataworking.*;
import com.interdataworking.mm.alg.*;

import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import edu.stanford.db.rdf.query.orderings.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;

// import edu.stanford.db.rdf.util.*;

public class LanguageCallbackSQL extends LanguageCallbackAdapter {

  SQLDDLVocabulary SQL = SQLDDLVocabulary.create();
  //  Cache cache = new Cache(100, false);

  QueryIterator[] Q_DEP = {
    new FindIterator("RC", SQL.foreignKeyRole, "FK"),
    new FindIterator("RC", SQL.uniqueKeyRole, "PK"),
    new FindIterator("FK", SQL.name, "FKN"),
    new FindIterator("FKT", SQL.column, "FK"),
    new FindIterator("FKT", SQL.name, "FKTN"),
    new FindIterator("PK", SQL.name, "PKN"),
    new FindIterator("PKT", SQL.column, "PK"),
    new FindIterator("PKT", SQL.name, "PKTN")
  };

  public QueryIterator[] Q_KEY_PAR_ = {
    new FindIterator("K", SQL.keyColumn, "C"),
    new FindIterator("C", SQL.name, "N")
  };

  public QueryIterator Q_KEY_PAR = new OrderIterator(new JoinIterator(Q_KEY_PAR_), new GroupBy("K"));

  public String toString() {

    return "Language callback for SQL DDL";
  }

  public boolean isCallbackFor(Model m) throws ModelException {

    return m.find(null, null, SQL.TableDef).size() > 0;
  }

  public Collection getConstraints() throws ModelException {

    List l = new ArrayList();
    int NONE = Merge.NONE;
    l.add(new Merge.CardinalityConstraint(SQL.name, NONE, 1));
    l.add(new Merge.CardinalityConstraint(SQL.type, NONE, 1));
    l.add(new Merge.CardinalityConstraint(SQL.keyColumn, 1, 1));
    l.add(new Merge.CardinalityConstraint(SQL.foreignKey, NONE, 1));
    l.add(new Merge.CardinalityConstraint(SQL.param, NONE, 1));
    l.add(new Merge.CardinalityConstraint(SQL.defaultValue, NONE, 1));
    l.add(new Merge.CardinalityConstraint(SQL.foreignKeyRole, NONE, 1));
    l.add(new Merge.CardinalityConstraint(SQL.uniqueKeyRole, NONE, 1));
    l.add(new Merge.CardinalityConstraint(SQL.rowOf, NONE, 1));
    l.add(new Merge.CardinalityConstraint(SQL.hasField, NONE, 1));

    l.add(new Merge.CardinalityConstraint(SQL.column, 1, NONE));
    l.add(new Merge.CardinalityConstraint(SQL.key, 1, NONE));
    l.add(new Merge.CardinalityConstraint(SQL.referentialConstraint, 1, NONE));

    return l;
  }

  public Set all(Model m) throws ModelException {

    Set result = new HashSet();
    for(Enumeration en = m.elements(); en.hasMoreElements();) {
      Statement s = (Statement)en.nextElement();
      if(!(s.subject() instanceof Statement))
	result.add(s.subject());
      if(!(s.object() instanceof Literal) &&
	 !(s.object() instanceof Statement) &&
	 !SQL.ALL.contains(s.object()))
	result.add(s.object());
    }
    return result;
  }

  public UntypedGateway defaultSerializer() throws ModelException {

    return new Obj2SQLDDL();
  }

  public List getDependencies(Model m) throws ModelException {

//     System.err.println("[LanguageCallbackSQL] computing dependencies of model " + m);

//     List l = (List)cache.get(m);
//     if(l != null) {
//       System.err.println("[LanguageCallbackSQL] reusing dependencies of model " + m);
//       return l;
//     }

    return new Select().getDependencies(m);
//     cache.put(m, l);

//     return l;
  }

  void addPath(Resource r, Map l2r, Set nodes, Set added, Collection stack) {

    //    System.err.println("-- add path: " + r + " at stack " + stack.size());

    if(!stack.add(r)) { // avoid cycles
      return; // already been here
    }

    if(stack.size() > 1 && nodes.contains(r))  { // yep!
	  
      // add all nodes along the path
      //      System.err.println("=== flushing stack on " + r + " : " + stack);
      added.addAll(stack);
    }

    Set s = (Set)l2r.get(r);

    if(s != null) {

      for(Iterator it = s.iterator(); it.hasNext();) {
	
	Resource t = (Resource)((MapPair)it.next()).getRight();
	addPath(t, l2r, nodes, added, stack);
      }
    }
    stack.remove(r);
  }

  public Model select(Model m, Set nodes) throws ModelException {

    // ADD ALL TABLES AND THEIR PK/FK THAT ARE ON THE PATH BETWEEN TWO TABLES

    Set checked = new HashSet(); // checked tables

    // find all table dependencies in m

    List tableDep = new ArrayList();

    {
      NodeIterator it = new NodeResultSet(Q_DEP, m);
      while(it.next()) {
	//	System.err.println("PKT: " + it.get("PKT") + " -> FKT: " + it.get("FKT"));
	//	System.err.println("   RC: " + it.get("RC"));
	// mutual references between tables
	tableDep.add(new MapPair(it.get("PKT"), it.get("RC")));
	tableDep.add(new MapPair(it.get("RC"), it.get("FKT")));
	tableDep.add(new MapPair(it.get("FKT"), it.get("RC")));
	tableDep.add(new MapPair(it.get("RC"), it.get("PKT")));
      }
    }

    Map l2r = MapPair.toJavaSetMap(tableDep, false);

    //    System.err.println("=== L2R: " + l2r);

    while(true) {

      Set added = new HashSet();
      
      for(Iterator it = nodes.iterator(); it.hasNext();) {

	Resource s = (Resource)it.next();
	addPath(s, l2r, nodes, added, new HashSet());
      }

      //      System.err.println("=== added elements on join paths: " + added);
      
      // now we need to collect all dependent nodes of the new nodes...
      
      Set requiredNodes = Reachable.reachable(added, getDependencies(m));

      //      System.err.println("=== elements required for join paths: " + requiredNodes);
      
      int oldCount = nodes.size();
      nodes.addAll(added);
      nodes.addAll(requiredNodes);
      
      if(oldCount == nodes.size())
	break;
    }

    // ADD UNIQUE KEY DEFINITIONS IF ALL FIELDS ARE SELECTED

    Set allColumns = new HashSet();
    Resource uk = null;

    Set addedKeys = new HashSet();

    for(NodeIterator it = new NodeResultSet(Q_KEY_PAR, m); it.next();) {

      //      System.err.println("== key:" + it.get("K") + ", col=" + it.get("C"));
      
      if(uk == null || !uk.equals(it.get("K"))) {

	if(uk != null && nodes.containsAll(allColumns)) {
	  //	  System.err.println("== adding unique key: " + uk);
	  addedKeys.add(uk);
	}
	allColumns.clear();
	uk = (Resource)it.get("K");
      }
      allColumns.add(it.get("C"));
    }

    // what about the last one?
    if(uk != null && nodes.containsAll(allColumns)) {
      //      System.err.println("== adding unique key: " + uk);
      addedKeys.add(uk);
    }

    if(!addedKeys.isEmpty()) { // add the keys and their dependent nodes
      
      nodes.addAll(addedKeys);
      nodes.addAll(Reachable.reachable(addedKeys, getDependencies(m)));
    }

    return new Subgraph().getSubgraph(m, nodes);
  }
}
