package com.interdataworking.gateway.sqlddl;

import java.io.*;
import java.util.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;
import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;
import org.w3c.rdf.vocabulary.dublin_core_19990702.DC;
import edu.stanford.db.rdf.query.orderings.*;
import com.interdataworking.mm.gui.TLineWriter;
import com.interdataworking.*;

public class Obj2SQLDDL implements RDFSerializer, UntypedGateway {

  TLineWriter w;

  static final String TAB = "    ";

  public SQLDDLVocabulary R = SQLDDLVocabulary.create();

  //  QueryIterator Q_TBL = new FindIterator("T", RDF.type, R.TableDef);

  public QueryIterator[] Q_TBL_ = {
    new FindIterator("T", RDF.type, R.TableDef),
    new FindIterator("T", R.name, "N")
  };

  public QueryIterator Q_TBL = new OrderIterator(new JoinIterator(Q_TBL_),
						 new Lexicographic("N"));

  public QueryIterator[] Q_COL_FOREIGN_ = {
    new FindIterator("RC", R.foreignKeyRole, "C"),
    new FindIterator("RC", R.uniqueKeyRole, "UC"),
    new FindIterator("UC", R.name, "UCN"),
    new FindIterator("UCT", R.column, "UC"),
    new FindIterator("UCT", R.name, "UCTN")
  };

  QueryIterator[] Q_COL_ = {
    new FindIterator("C", RDF.type, R.ColDef),
    new FindIterator("T", R.column, "C", "S"),
    new FindIterator("C", R.name, "N"),
    new FindIterator("C", R.type, "CT"),
    new FindIterator("CT", R.name /*RDF.value*/, "V"),
    new OptionalIterator(new FindIterator("C", DC.Description, "DESC")),
    new OptionalIterator(new FindIterator("C", R.defaultValue, "D")),
    new OptionalIterator(new FindIterator("C", R.constraint, "CC")),
    new OptionalIterator(new JoinIterator(Q_COL_FOREIGN_))
  };
  //  QueryIterator Q_COL = new OrderIterator(new JoinIterator(Q_COL_), "S");
  public QueryIterator Q_COL = new OrderIterator(new UniqueIterator(new JoinIterator(Q_COL_), "C"),
						 new ByOrderProperty("S"));

  public QueryIterator Q_PAR =
    new OrderIterator(new FindIterator("CT", R.param, "P", "S"),
		      new ByOrderProperty("S"));

  public QueryIterator[] Q_KEY = {
    new FindIterator("T", R.key, "K"),
    new FindIterator("K", RDF.type, "KT"),
    new OptionalIterator(new FindIterator("K", R.name, "N"))
    //    new FindIterator("K", R.name, "N", FindIterator.OPTIONAL)
  };

  public QueryIterator[] Q_KEY_PAR_ = {
    new FindIterator("K", R.keyColumn, "C", "S"),
    new FindIterator("C", R.name, "N")
  };
  public QueryIterator Q_KEY_PAR =
    new OrderIterator(new JoinIterator(Q_KEY_PAR_),
		      new ByOrderProperty("S"));


  public Resource findTable(Model model, String name) throws ModelException {

    NodeIterator it = new NodeResultSet(Q_TBL_, model);
    it.set("N", model.getNodeFactory().createLiteral(name));

    if(it.next())
      return (Resource)it.get("T");
    else
      return null;
  }

  public Resource findField(Model model, String tableName, String fieldName) throws ModelException {

    Resource table = findTable(model, tableName);
    if(table == null)
      return null;

    NodeIterator it = new NodeResultSet(Q_COL_, model);
    it.set("T", table);
    it.set("N", model.getNodeFactory().createLiteral(fieldName));

    if(it.next())
      return (Resource)it.get("C");
    else
      return null;
  }

  public void serialize(Model model, Writer out) throws SerializationException, IOException, ModelException {

    //    System.err.println("Num: " + model.find(null, RDF.type, R.TableDef).size());

    this.w = new TLineWriter(out);

    NodeIterator it = new NodeResultSet(Q_TBL, model);
    w.write("# Table definitions\n\n");
    while(it.next()) {
      w.write("CREATE TABLE ");
      w.write(String.valueOf(it.get("N")), it.get("T"));
      w.write(" (\n");

      // write column definition

      String description = null; // optional column description

      NodeIterator itc = new NodeResultSet(Q_COL, model);
      itc.set("T", it.get("T"));
      boolean first_itc = true;
      while(itc.next()) {

	if(!first_itc) {
	  w.write(",");
	  if(description != null) {
	    w.write(TAB + "# " + description);
	    description = null;
	  }
	  w.write("\n");
	}
	first_itc = false;
	
	w.write(TAB);
	w.write(String.valueOf(itc.get("N")), itc.get("C"));
	w.write(" ");
	w.write(String.valueOf(itc.get("V"))); // no node for type, itc.get("CT"));

	// column definition parameters
	NodeIterator itp = new NodeResultSet(Q_PAR, model);
	itp.set("CT", itc.get("CT"));
	boolean first_itp = true;
	if(itp.next()) {
	  w.write("(");
	  do {
	    if(!first_itp)
	      w.write(",");
	    w.write("" + itp.get("P"));
	    first_itp = false;
	  } while(itp.next());
	  w.write(")");
	}

	// description
	Literal dlit = (Literal)itc.get("DESC");
	description = dlit != null ? dlit.getLabel() : null;

	// default value
	if(itc.get("D") != null)
	  w.write(" DEFAULT " + itc.get("D"));

	// NOT NULL
	if(itc.get("CC") != null)
	  w.write(" NOT NULL");
	//	w.write('\n');

	// REFERENCES
	if(itc.get("UC") != null) {

	  // obtain table name of the referenced column
	  w.write(" ");
	  w.write("REFERENCES", itc.get("RC"));
	  w.write(" ");
	  w.write(String.valueOf(itc.get("UCTN")));
	  RDFNode thatColName = (RDFNode)itc.get("UCN");
	  if(!itc.get("N").equals(thatColName))
	    w.write("(" + thatColName + ")");
	}
      }

      String ksep = (description != null) ? "," + TAB + "# " + description + "\n" : ",\n";

      // WRITE KEY DEFINITIONS
      NodeIterator itk = new NodeResultSet(Q_KEY, model);
      itk.set("T", it.get("T"));
      while(itk.next()) {

	w.write(ksep);
	ksep = ",\n"; description = null;
	// determine key type
	Object ktObj = itk.get("KT");
	if(!(ktObj instanceof Resource))
	   System.err.println("Error: key type is not a resource: " + ktObj);
	Resource kt = (Resource)ktObj;
	String ktstr = "KEY";
	if(R.PrimaryKey.equals(kt))
	  ktstr = "PRIMARY KEY";
	else if(R.UniqueKey.equals(kt))
	  ktstr = "UNIQUE";

	w.write(TAB);
	w.write(ktstr, itk.get("K"));

	String keyName = itk.get("N") != null ? itk.get("N").getLabel() : null;
	if(keyName != null)
	  w.write(" " + keyName);

	// key components
	NodeIterator itkc = new NodeResultSet(Q_KEY_PAR, model);
	itkc.set("K", itk.get("K"));
	boolean first_itkc = true;
	if(itkc.next()) {
	  w.write(" (");
	  do {
	    if(!first_itkc)
	      w.write(",");
	    w.write("" + itkc.get("N"));
	    first_itkc = false;
	  } while(itkc.next());
	  w.write(")");
	}
      }

      if(description != null) {
	w.write(TAB + "# " + description);
	description = null;
      }
      w.write("\n);\n");

    }
    w.flush();
  }


  /**
   * Transforms an ordered list of models into another list of models
   */
  public List execute(List input) throws ModelException {

    Model m = (Model)input.get(0);
    StringWriter w = new StringWriter();
    try {
      serialize(m, w);
    } catch (Exception any) {}

    List result = new ArrayList();
    result.add(w.toString());
    result.add(this.w.getLines());

    //    System.err.println("Returning string:\n " + w + ", lines:\n" + lines);

    return result;
  }

  public int getMinInputLen() { return 1; }

  public int getMaxInputLen() { return 1; }

  public int getMinOutputLen() { return 2; }

  public int getMaxOutputLen() { return 2; }

}
