package com.interdataworking.gateway.sqlddl;

import java.io.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;
import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;
import org.w3c.rdf.vocabulary.dublin_core_19990702.DC;
import edu.stanford.db.rdf.query.orderings.*;

public class Obj2XSD extends Obj2SQLDDL {

  public void serialize(Model model, Writer w) throws SerializationException, IOException, ModelException {

    //    System.err.println("Num: " + model.find(null, RDF.type, R.TableDef).size());

    NodeIterator it = new NodeResultSet(Q_TBL, model);
    w.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n");
    w.write("<xsd:schema xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\">\n");

    while(it.next()) {
      w.write("\t<xsd:complexType name=\"" + it.get("N") + "\">\n");
      w.write("\t<xsd:sequence>\n");

      // write column definition

      String description = null; // optional column description

      NodeIterator itc = new NodeResultSet(Q_COL, model);
      itc.set("T", it.get("T"));
      boolean first_itc = true;
      while(itc.next()) {

	if(!first_itc) {
	  //	  w.write(",");
	  if(description != null) {
	    w.write("\t\t<!-- " + description + " -->");
	    description = null;
	  }
	  //	  w.write("\n");
	}
	first_itc = false;
	

	String elementType = "";

	// REFERENCES
	if(itc.get("UC") != null) {

	  // obtain table name of the referenced column

	  elementType = String.valueOf(itc.get("UCTN"));

	  /*
	  w.write(" REFERENCES " + itc.get("UCTN"));
	  RDFNode thatColName = (RDFNode)itc.get("UCN");
	  if(!itc.get("N").equals(thatColName))
	    w.write("(" + thatColName + ")");
	  */
	} else { // regular field

	  elementType = String.valueOf(itc.get("V"));

	  String e = elementType.toLowerCase();
	  if(e.startsWith("int"))
	    elementType = "xsd:int";
	  else if(e.startsWith("string") || e.startsWith("char") || e.startsWith("varchar"))
	    elementType = "xsd:string";
	  else if(e.startsWith("dec") || e.startsWith("double") || e.startsWith("num"))
	    elementType = "xsd:real";
	  else if(e.startsWith("date"))
	    elementType = "xsd:date";
	}

	w.write("\t\t<xsd:element name=\"" + itc.get("N") + "\" type=\"" + elementType + "\"");

	/*
	// column definition parameters
	NodeIterator itp = new NodeResultSet(Q_PAR, model);
	itp.set("CT", itc.get("CT"));
	boolean first_itp = true;
	if(itp.next()) {
	  w.write("(");
	  do {
	    if(!first_itp)
	      w.write(',');
	    w.write("" + itp.get("P"));
	    first_itp = false;
	  } while(itp.next());
	  w.write(')');
	}
	*/

	// description
	Literal dlit = (Literal)itc.get("DESC");
	description = dlit != null ? dlit.getLabel() : null;

	// default value
	if(itc.get("D") != null)
	  w.write(" default=\"" + itc.get("D") + "\"");
	
	// NOT NULL
	if(itc.get("CC") != null)
	  w.write(" use=\"required\"");
	//	  w.write(" NOT NULL");
	//	w.write('\n');

	w.write("/>\n");
      }
      
      /*
      String ksep = (description != null) ? ",\t# " + description + "\n" : ",\n";

      // WRITE KEY DEFINITIONS
      NodeIterator itk = new NodeResultSet(Q_KEY, model);
      itk.set("T", it.get("T"));
      while(itk.next()) {

	w.write(ksep);
	ksep = ",\n"; description = null;
	// determine key type
	Object ktObj = itk.get("KT");
	if(!(ktObj instanceof Resource))
	   System.err.println("Error: key type is not a resource: " + ktObj);
	Resource kt = (Resource)ktObj;
	String ktstr = "KEY";
	if(R.PrimaryKey.equals(kt))
	  ktstr = "PRIMARY KEY";
	else if(R.UniqueKey.equals(kt))
	  ktstr = "UNIQUE KEY";

	w.write("\t" + ktstr);

	String keyName = itk.get("N") != null ? itk.get("N").getLabel() : null;
	if(keyName != null)
	  w.write(" " + keyName);

	// key components
	NodeIterator itkc = new NodeResultSet(Q_KEY_PAR, model);
	itkc.set("K", itk.get("K"));
	boolean first_itkc = true;
	if(itkc.next()) {
	  w.write(" (");
	  do {
	    if(!first_itkc)
	      w.write(',');
	    w.write("" + itkc.get("N"));
	    first_itkc = false;
	  } while(itkc.next());
	  w.write(')');
	}
      }
      */

      if(description != null) {
	w.write(" <-- " + description + " -->\n");
	description = null;
      }

      w.write("\t</xsd:sequence>\n");
      w.write("\t</xsd:complexType>\n");
    }

    w.write("</xsd:schema>\n");
    w.flush();
  }
}
