package com.interdataworking.gateway.sqlview;

import java.util.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import com.interdataworking.*;
import com.interdataworking.mm.alg.*;
import com.interdataworking.gateway.sqlddl.Obj2SQLDDL;
import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.query.orderings.*;
import edu.stanford.db.rdf.util.*;

public class ExtractMap implements UntypedGateway {

  Obj2SQLDDL sqlddl = new Obj2SQLDDL();

  public static void collectFields(Resource t, Model m, Collection result) throws ModelException {

    if(RDFUtil.isInstanceOf(m, t, SQLViewVocabulary.FieldTerm)) {

      Resource ref = RDFUtil.getObjectResource(m, t, SQLViewVocabulary.fieldRef);
      if(ref != null)
	result.add(ref);
      return;
    }

    for(Enumeration en = m.find(t, null, null).elements(); en.hasMoreElements();) {
	
      Statement st = (Statement)en.nextElement();
      if(Text2Ref.SELECT_PARENT_ARCS.contains(st.predicate()))
	if(st.object() instanceof Resource)
	  collectFields((Resource)st.object(), m, result);
    }
  }

  public List extractMap(Model m_) throws ModelException {

    PowerModel m = PowerModel.create(m_);
    List map = new ArrayList();

    for(NodeIterator iv = new NodeResultSet(Text2Ref.Q_VIEW_, m); iv.next();) {

      Resource view = (Resource)iv.get("V");
      Resource select = (Resource)iv.get("S");

      NodeIterator it = new NodeResultSet(Text2Ref.Q_VIEW_FIELDS_ONLY_ORDERED, m);
      it.set("V", view);
      NodeIterator is = m.getOrderedNI(select, SQLViewVocabulary.selectTerm);

      while(it.next() && is.next()) {

	Resource target = (Resource)it.get("F");
	Resource term = (Resource)is.get("O");

	//	System.err.println("Term " + term + " -> " + target);

	Set fields = new HashSet();
	collectFields(term, m, fields);

	for(Iterator ic = fields.iterator(); ic.hasNext();) {
	  //	  System.err.println("  Field " + ic.next() + " -> " + target);
	  map.add(new MapPair((RDFNode)ic.next(), target));
	}
      }
    }

    return map;
  }

  public List execute(List input) throws ModelException {

    Model m = (Model)input.get(0);

    List map = extractMap(m);

    List result = new ArrayList();
    result.add(MapPair.asModel(m, map));
    return result;
  }

  public int getMinInputLen() { return 1; }

  public int getMaxInputLen() { return 1; }

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }
}
