package com.interdataworking.gateway.xml;

import java.io.*;
import java.util.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;
import org.w3c.rdf.util.*;

import com.interdataworking.*;
import com.interdataworking.mm.alg.*;

import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import edu.stanford.db.rdf.query.orderings.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;

// import edu.stanford.db.rdf.util.*;

public class LanguageCallbackXML extends LanguageCallbackAdapter {

  public XMLVocabulary R = new XMLVocabulary();

  public String toString() {

    return "Language callback for XML and XSD";
  }

  public boolean isCallbackFor(Model m) throws ModelException {

    return m.find(null, null, R.XMLDOM).size() > 0;
  }

  public Collection getConstraints() throws ModelException {

    List l = new ArrayList();
    int NONE = Merge.NONE;

    com.interdataworking.gateway.xml.XMLVocabulary XML = com.interdataworking.gateway.xml.XMLVocabulary.create();

    l.add(new Merge.CardinalityConstraint(XML.child, 1, NONE));

    l.add(new Merge.CardinalityConstraint(XML.tagName, NONE, 1));
    l.add(new Merge.CardinalityConstraint(XML.cdata, NONE, 1));

    com.interdataworking.gateway.xml.XSDVocabulary XSD = new com.interdataworking.gateway.xml.XSDVocabulary();

    l.add(new Merge.CardinalityConstraint(XSD.name, NONE, 1));
    l.add(new Merge.CardinalityConstraint(XSD.type, NONE, 1));
    l.add(new Merge.CardinalityConstraint(XSD.ref, NONE, 1));
    l.add(new Merge.CardinalityConstraint(XSD.minOccurs, NONE, 1));
    l.add(new Merge.CardinalityConstraint(XSD.maxOccurs, NONE, 1));
    l.add(new Merge.CardinalityConstraint(XSD.use, NONE, 1));
    l.add(new Merge.CardinalityConstraint(XSD.default_, NONE, 1));
    l.add(new Merge.CardinalityConstraint(XSD.base, NONE, 1));
    l.add(new SingleRootConstraint());

    return l;
  }

  class SingleRootConstraint implements Merge.Constraint {

    Merge.Constraint card = new Merge.CardinalityConstraint(RDF.type, 1, 1);

    public SingleRootConstraint() {
    }

    // predicate that triggers validation of this constraint (or null if not tested)
    public Collection getPredicateTriggers() {

      List l = new ArrayList();
      l.add(RDF.type);
      return l;
    }

    // would the model still be ok if the statement st is added?
    public boolean isSatisfied(Statement st, Model result) throws ModelException {

      if(R.XMLDOM.equals(st.object()))
	return card.isSatisfied(st, result);
      return true;
    }

    public String toString() {

      return "XMLDOM single root constraint";
    }
  }

  public UntypedGateway defaultSerializer() throws ModelException {

    return new RDF2XML2();
  }

  public Set all(Model m) throws ModelException {

    Set result = new HashSet();
    for(Enumeration en = m.elements(); en.hasMoreElements();) {
      Statement s = (Statement)en.nextElement();
      if(!(s.subject() instanceof Statement))
	result.add(s.subject());
      if(s.object() instanceof Resource &&
	 !(s.object() instanceof Statement) &&
	 !XMLVocabulary.ALL.contains(s.object()) && 
	 !XSDVocabulary.ALL.contains(s.object()) &&
	 !XMLVocabulary.usesNamespace.equals(s.predicate()) &&
	 !XMLVocabulary.tagName.equals(s.predicate()))
	result.add(s.object());
    }
    return result;
  }

  public List getDependencies(Model m) throws ModelException {

    return new DependenciesXSD().getDependencies(m);
  }

  public Model select(Model m, Set nodes) throws ModelException {

    return new Subgraph().getSubgraph(m, nodes);
  }

}
