package com.interdataworking.mm;

import java.util.*;
import java.io.*;
import org.w3c.rdf.syntax.*;
import org.w3c.rdf.model.*;
// import com.interdataworking.mm.alg.MapVocabulary;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import org.xml.sax.InputSource;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;

public class EvaluateMapping implements RDFParser {

  String ns = "urn:literal:";

  public void parse(InputSource source, RDFConsumer consumer) throws SAXException {

    try {
      consumer.startModel();

      Set res = toItems(source.getCharacterStream());
      NodeFactory f = consumer.getNodeFactory();
      MapVocabulary M = new MapVocabulary(f);

      Iterator it = res.iterator();
      while(it.hasNext()) {

	String item = (String)it.next();
	int p = item.indexOf(' ');
	Resource r1 = f.createResource(ns, item.substring(0, p));
	Resource r2 = f.createResource(ns, item.substring(p+1));
	Resource m = f.createUniqueResource();

	consumer.addStatement(f.createStatement(m, RDF.type, M.Map));
	consumer.addStatement(f.createStatement(m, M.src, r1));
	consumer.addStatement(f.createStatement(m, M.dest, r2));
	consumer.addStatement(f.createStatement(m, M.sim, f.createLiteral("1.0")));
      }

      consumer.endModel();

    } catch(Exception ioex) {

      throw new SAXException(ioex);
    }
  }

  public void setErrorHandler (ErrorHandler handler) {}
  

  static Set toItems(String fname) throws IOException {

    return toItems(new FileReader(fname));
  }

  static Set toItems(Reader reader) throws IOException {

    Set result = new HashSet();

    BufferedReader in = new BufferedReader(reader);

    String line = null;
    while((line = in.readLine()) != null) {

      StringTokenizer st = new StringTokenizer(line, " \t\n\r");
      if(st.hasMoreTokens()) {
	String s1 = st.nextToken();
	String s2 = st.nextToken();
	String item = s1 + " " + s2;
	result.add(item);
      }
    }

    return result;
  }


  public static void main(String args[]) throws Exception {

    if(args.length < 2) {
      System.err.println("Usage: EvaluateMapping {<test map> <correct map>}*");
      System.exit(1);
    }

    int totalSaving = 0;
    int numTests = 0;
    int i = 0;
    while(i + 1 < args.length) {

      numTests++;

      Set actual = toItems(args[i]);
      Set intended = toItems(args[i+1]);

      int correctNum = 0;
    
      Iterator it = actual.iterator();
      while(it.hasNext()) {

	Object p = it.next();
	if(intended.contains(p))
	  correctNum++;
      }

      int subOp = (actual.size() - correctNum);
      int addOp = (intended.size() - correctNum);

      int accuracy = (100 - 100 * (subOp+addOp) / intended.size());

      String res = "Recall: " + (100 * correctNum / intended.size()) + "%, " +
	"Precision: " + (100 * correctNum / actual.size()) + "% " +
	"(computed: " + actual.size() + ", computed correct: " + correctNum +
	", intended: " + intended.size() + ") " +
	"-" + subOp + " +" + addOp +
	" => " + (subOp + addOp) + " op, " +
	"accuracy (saving): " + accuracy + "%";

      if(accuracy > 0)
	totalSaving += accuracy;

      System.out.println(res);
      i += 2;
    }

    System.err.println("Avg savings over " + numTests + " tests: " + (totalSaving / numTests) + "%");
  }
}
