package com.interdataworking.mm.alg;

import java.sql.*;
import java.util.*;

public class SimFloodSQL {

  static final boolean DEBUG = true;

  static final boolean STRAIGHT_FLOW_JOIN = true;
  static final boolean PRESERVE_TABLES_FOR_DEBUG = true;
  static final double MIN_FLOW_SIM = 0.1;

  static String JDBC_DRIVER_NAME = "org.gjt.mm.mysql.Driver";
  static String JDBC_URL = "jdbc:mysql://localhost/simflood";
  static String JDBC_USER_NAME = "melnik";
  static String JDBC_PASSWD = "";

  /*
  static String JDBC_DRIVER_NAME = "org.gjt.mm.mysql.Driver";
  static String JDBC_URL = "jdbc:mysql://Skeptic.Stanford.EDU/simflood";
  static String JDBC_USER_NAME = "repository";
  static String JDBC_PASSWD = "";
  */

  public static final int TBL_SP_CARD = 1;
  public static final int TBL_SPO = 2;
  public static final int TBL_NODES = 3;
  public static final int TBL_SIGMA = 4;

  public String tableNamePrefix = ""; //"001_";
  public String BIGINT = "bigint";
  public String INT = "int";
  public String DOUBLE = "double";

  public String T_TI;
  public String T_MAP_PAIRS;

  public Connection con = null;

  Set createdTables = new HashSet();

  public SimFloodSQL() throws SQLException {

    // create map pairs

    T_TI = tableName("ti");
    T_MAP_PAIRS = tableName("mapPairs");

    String mapPairs = T_MAP_PAIRS;
    createTable(mapPairs, TBL_SIGMA);
  }

  public Statement createStatement() throws SQLException {

    if(con == null)
      con = getConnection();
    return con.createStatement();
  }

  public Connection getConnection() throws SQLException {

    try {
      Class.forName(JDBC_DRIVER_NAME);
    } catch (Exception any) {
      System.err.println("Could not load JDBC driver: " + JDBC_DRIVER_NAME);
    }

    Connection con = (!"".equals(JDBC_USER_NAME)) ?
      DriverManager.getConnection(JDBC_URL, JDBC_USER_NAME, JDBC_PASSWD) :
      DriverManager.getConnection(JDBC_URL);

    return con;
  }

  public void createSPOIndex(String tname) throws SQLException {

    execute("CREATE INDEX " + tname+"_s ON " + tname + " (s)");
    execute("CREATE INDEX " + tname+"_p ON " + tname + " (p)");
    execute("CREATE INDEX " + tname+"_o ON " + tname + " (o)");
  }

  public void createCardIndex(String tname) throws SQLException {

    execute("CREATE UNIQUE INDEX " + tname+"_sp ON " + tname + "(s,p)");
    execute("CREATE INDEX " + tname+"_p ON " + tname + "(p,s)");
  }

  public void createSigmaIndex(String tname) throws SQLException {

    execute("CREATE UNIQUE INDEX " + tname+"_node_ab ON " + tname + "(node_a,node_b)");
    execute("CREATE UNIQUE INDEX " + tname+"_node_ba ON " + tname + "(node_b,node_a)");
  }


//   public void createLaxSigmaIndex(String tname) throws SQLException {

//     execute("CREATE INDEX " + tname+"_node ON " + tname + "(node_a,node_b)");
//   }


  public void createTable(String tname, int ttype) throws SQLException {

    createdTables.add(tname);

    String s = null;
    //    String CT = DEBUG ? "CREATE TABLE " : "CREATE TEMPORARY TABLE ";
    String CT = "CREATE TABLE ";

    switch(ttype) {

    case TBL_SP_CARD:
      s = CT + tname + " (" +
	"s " + BIGINT + " NOT NULL," +
	"p " + BIGINT + " NOT NULL," +
	"card " + INT + " NOT NULL" +
	")";
      break;

    case TBL_SPO:
      s = CT + tname + " (" +
	"s " + BIGINT + " NOT NULL," +
	"p " + BIGINT + " NOT NULL," +
	"o " + BIGINT + " NOT NULL" +
	")";
      break;

    case TBL_NODES:
      s = CT + tname + " (" +
	"node " + BIGINT + " NOT NULL PRIMARY KEY" +
	")";
      break;

    case TBL_SIGMA:
      s = CT + tname + " (" +
	"node_a " + BIGINT + " NOT NULL," +
	"node_b " + BIGINT + " NOT NULL," +
	"sim " + DOUBLE + " DEFAULT 0" +
	")";
      break;
    }

    if(s == null)
      throw new RuntimeException("Unknown table type: " + ttype);

    dropTable(tname);
    execute(s);
  }

  public void dropTable(String tname) {

    try {
      execute("DROP TABLE " + tname);
    } catch (SQLException any) {}
  }


  public String tableName(String name) {

    return tableNamePrefix + name;
  }

  public void execute(String s) throws SQLException {

    if(con == null)
      con = getConnection();

    Statement stmt = con.createStatement();
    if(DEBUG)
      System.err.println("EXEC: " + s);
    stmt.execute(s);
  }

  public ResultSet executeQuery(String s) throws SQLException {

    if(con == null)
      con = getConnection();

    Statement stmt = con.createStatement();
    if(DEBUG)
      System.err.println("Q: " + s);
    return stmt.executeQuery(s);
  }


  public void computeCard(String srcTable, String s, String p, String destTable) throws SQLException {

    s = srcTable + "." + s;
    p = srcTable + "." + p;

    String q = "INSERT INTO " + destTable + " " +
      "SELECT " + s + ", " + p + ", COUNT(*)  " +
      "FROM " + srcTable + " " +
      "GROUP BY " + s + ", " + p;

    execute(q);
  }

  public void normalize(String t, String sim) throws SQLException {

    sim = t + "." + sim;

    String q = "SELECT MAX(" + sim + ") " + 
      "FROM " + t;

    ResultSet rs = executeQuery(q);
    if(rs.next()) {

      double max = rs.getDouble(1);
      if(max > 0) {
	q = "UPDATE " + t + " SET " + sim + "=" + sim + "/" + max;
	execute(q);
      }
    } // else do nothing

  }


  // all values in t1 and t2 are guaranteed to be preserved
  // since MySQL does not support outer join, implement it ourselves
  public void addSim(String dest, String t1, String t2) throws SQLException {

    String t1_node_a = t1 + ".node_a";
    String t1_node_b = t1 + ".node_b";
    String t2_node_a = t2 + ".node_a";
    String t2_node_b = t2 + ".node_b";
    String t1_sim = t1 + ".sim";
    String t2_sim = t2 + ".sim";

    String q = "INSERT INTO " + dest + " " + 
      "SELECT " + t1_node_a + ", " + t1_node_b + ", IF(ISNULL(" + t2_sim + "), " + t1_sim + ", (" + t1_sim + "+ " + t2_sim + ")) " +
      "FROM " + t1 + " LEFT JOIN " + t2 + " ON " +
      t1_node_a + "=" + t2_node_a + " AND " +
      t1_node_b + "=" + t2_node_b;

    execute(q);

    q = "INSERT INTO " + dest + " " + 
      "SELECT " + t2_node_a + "," + t2_node_b + ", " + t2_sim + " " +
      "FROM " + t1 + " RIGHT JOIN " + t2 + " ON " +
      t1_node_a + "=" + t2_node_a + " AND " +
      t1_node_b + "=" + t2_node_b + " " +
      "WHERE " + t1_node_a + " is NULL ";

    execute(q);

  }


  public void copySim(String src, String dest) throws SQLException {

    String src_node_a = src + ".node_a";
    String src_node_b = src + ".node_b";
    String src_sim = src + ".sim";

    String q = "INSERT INTO " + dest + " " + 
      "SELECT " + src_node_a + ", " + src_node_b + ", " + src_sim + " " +
      "FROM " + src;

    execute(q);
  }

  public void addToMapPairs(String t, String n1, String n2,
			    String mapPairs) throws SQLException {

    n1 = t + "." + n1;
    n2 = t + "." + n2;

    String mapPairs_n1 = mapPairs + ".n1";
    String mapPairs_n2 = mapPairs + ".n2";

    String q = "INSERT INTO " + mapPairs + " (" + mapPairs_n1 + "," + mapPairs_n2 + ") " +
      "SELECT " + n1 + "," + n2 + " " +
      "FROM " + t;

    execute(q);
  }


  // we don't lose values from src
  public void flow(String dest, String src, String A, String B, String s, String p, String o, String cardA_Inv, String cardB_Inv) throws SQLException {

    // s,s from o,o

    /*
      select A.o, B.o, sum(if(isNull(S.sim),0,S.sim) * 2 / (cardA.card+cardB.card)) AS newSim from src AS A, dest AS B, cardA, cardB left join tiSum AS S ON S.node_a=A.s AND S.node_b=B.s where A.p=B.p and A.s=cardA.s and A.p=cardA.p and B.s=cardB.s and B.p=cardB.p group by A.o, B.o having newSim > 0;
    */


    String A_s = "A." + s;
    String A_o = "A." + o;
    String A_p = "A." + p;
    String B_s = "B." + s;
    String B_o = "B." + o;
    String B_p = "B." + p;

    String cardA_Inv_s = cardA_Inv + ".s";
    String cardA_Inv_p = cardA_Inv + ".p";
    String cardA_Inv_card = cardA_Inv + ".card";

    String cardB_Inv_s = cardB_Inv + ".s";
    String cardB_Inv_p = cardB_Inv + ".p";
    String cardB_Inv_card = cardB_Inv + ".card";

    String SELECT = STRAIGHT_FLOW_JOIN ? "SELECT STRAIGHT_JOIN" : "SELECT";

    String q = "insert into " + dest + " " +
      SELECT + " A.s, B.s, " +
      getPiFormula(cardA_Inv_card, cardB_Inv_card) +
      "from " + src + " AS S, " +
      A + " AS A, " + B + " AS B, " +
      cardA_Inv + ", " + cardB_Inv +
      " where S.node_a=" + A_o + " and S.node_b=" + B_o + " and " +
      A_p + "=" + B_p + " and " + A_o + "=" + cardA_Inv_s +
      " and " + A_p + "=" + cardA_Inv_p + " and " + B_o + "=" + cardB_Inv_s + " and " +
      B_p + "=" + cardB_Inv_p + " group by " + A_s + ", " + B_s + " having newSim > " + MIN_FLOW_SIM;

    execute(q);
  }

  String getPiFormula(String cardA_card, String cardB_card) {

    return "sum(if(isNull(S.sim),0,S.sim) * 2 / (" + cardA_card + "+" + cardB_card + ")) AS newSim ";
  }

  // we don't lose values from src
  public void flowInv(String dest, String src, String A, String B, String s, String p, String o, String cardA, String cardB) throws SQLException {

    // o,o from s,s

    String A_s = "A." + s;
    String A_o = "A." + o;
    String A_p = "A." + p;
    String B_s = "B." + s;
    String B_o = "B." + o;
    String B_p = "B." + p;

    String cardA_s = cardA + ".s";
    String cardA_p = cardA + ".p";
    String cardA_card = cardA + ".card";

    String cardB_s = cardB + ".s";
    String cardB_p = cardB + ".p";
    String cardB_card = cardB + ".card";

    String SELECT = STRAIGHT_FLOW_JOIN ? "SELECT STRAIGHT_JOIN" : "SELECT";

    String q = "INSERT INTO " + dest + " " +
      SELECT + " A.o, B.o, " +
      getPiFormula(cardA_card, cardB_card) +
      "from " + src + " AS S, " +
      A + " AS A, " + B + " AS B, " +
      cardA + ", " + cardB +
      " where S.node_a=" + A_s + " and S.node_b=" + B_s + " and " +
      A_p + "=" + B_p + " and " + A_s + "=" + cardA_s +
      " and " + A_p + "=" + cardA_p + " and " + B_s + "=" + cardB_s + " and " +
      B_p + "=" + cardB_p + " group by " + A_o + ", " + B_o + " having newSim > " + MIN_FLOW_SIM;

    /*

    select A.o, B.o, sum(if(isNull(S.sim),0,S.sim) * 2 / (cardA.card+cardB.card)) AS newSim
    from src AS A, dest AS B,
    cardA, cardB left join tiSum AS S ON S.node_a=A.s
    AND S.node_b=B.s where A.p=B.p and A.s=cardA.s
    and A.p=cardA.p and B.s=cardB.s and
    B.p=cardB.p group by A.o, B.o having newSim > 0;

    */

    execute(q);
  }

//   public double residualLen(String t1, String t2) throws SQLException {
//   }

  public ResultSet getResult(String t, String mapPairs) throws SQLException {

//     String mapPairs_n1 = mapPairs + ".node_a";
//     String mapPairs_n2 = mapPairs + ".node_b";
//     String mapPairs_node = mapPairs + ".node";

    String t_node_a = t + ".node_a";
    String t_node_b = t + ".node_b";
    String t_sim = t + ".sim";

    String q = "SELECT " + t_node_a + "," + t_node_b + "," + t_sim + " " +
      "FROM " + t;

    return executeQuery(q);
  }

  public ResultSet getResult() throws SQLException {

    return getResult(T_TI, T_MAP_PAIRS);
  }

  public String getMapPairsTableName() {

    return T_MAP_PAIRS;
  }

  public String getSigmaTableName() {

    return T_TI;
  }

  public void cleanUp() throws SQLException {

    Iterator it = createdTables.iterator();
    while(it.hasNext()) {
      String tname = (String)it.next();
      dropTable(tname);
    }
  }

  public void addToNodes(String t, String field, String dest) throws SQLException {

    field = t + "." + field;

    String q = "INSERT INTO " + dest + " " +
      "SELECT DISTINCT " + field + " FROM " + t;
    
    execute(q);
  }

  public void defaultMapPairs(String src, String dest) throws SQLException {

    defaultMapPairs(src, dest, "s", "p", "o");
  }

  public void defaultMapPairs(String src, String dest, String s, String p, String o) throws SQLException {

    String A_nodes = tableName("A_nodes");
    createTable(A_nodes, TBL_NODES);

    addToNodes(src, s, A_nodes);
    addToNodes(src, o, A_nodes);

    String B_nodes = tableName("B_nodes");
    createTable(B_nodes, TBL_NODES);

    addToNodes(dest, s, B_nodes);
    addToNodes(dest, o, B_nodes);

    String mapPairs = T_MAP_PAIRS;
    String q = "INSERT INTO " + mapPairs + " (node_a,node_b,sim)" +
      "SELECT A.node,B.node,1.0 " +
      "FROM " + A_nodes + " AS A, " + B_nodes + " AS B "; // +
    //      "WHERE A.node=B.node";

    execute(q);

    if(!PRESERVE_TABLES_FOR_DEBUG) {
      dropTable(A_nodes);
      dropTable(B_nodes);
    }
  }

  /*
  public void defaultMapPairs(String pairs, String src, String dest, String s, String p, String o) throws SQLException {

    // equal => 1, otherwise: 0.001
    addToMapPairsColumn(pairs, src, dest, s, s);
    addToMapPairsColumn(pairs, src, dest, o, o);
    addToMapPairsColumn(pairs, src, dest, s, o);
    addToMapPairsColumn(pairs, src, dest, o, s);
  }

  public void addToMapPairsColumn(String mapPairs, String src, String dest, String srcCol, String destCol) throws SQLException {

    srcCol = "A." + srcCol;
    destCol = "B." + destCol;

    String mapPairs_n1 = mapPairs + ".node_a";
    String mapPairs_n2 = mapPairs + ".node_b";
    String mapPairs_sim = mapPairs + ".sim";

    String q = "INSERT INTO " + mapPairs +
      " (" + mapPairs_n1 + "," + mapPairs_n2 + "," + mapPairs_sim + ") " +
      "SELECT DISTINCT " + srcCol + ", " + destCol + ", ";

    boolean equalOnly = true;

    if(equalOnly) {
      q += "1.0 " +
	"FROM " + src + " AS A," + dest + " AS B " + 
	"WHERE " + srcCol + "=" + destCol;
    } else {
      q += "IF(" + srcCol + "=" + destCol + ", 1.0, 0.001) " +
	"FROM " + src + " AS A," + dest + " AS B";
    }

    execute(q);
  }
  */

  public void mjoin(String src, String dest, int iterations) throws SQLException {

    mjoin(src, dest, "s", "p", "o", iterations);
  }

  public void mjoin(String src, String dest, String s, String p, String o, int iterations) throws SQLException {

    String srcCard = tableName("cardA");
    String destCard = tableName("cardB");

    createTable(srcCard, TBL_SP_CARD);
    computeCard(src, s, p, srcCard);

    createTable(destCard, TBL_SP_CARD);
    computeCard(dest, s, p, destCard);

    // create indexes on card
    createCardIndex(destCard);
    createCardIndex(srcCard);

    String srcCardInv = tableName("cardA_Inv");
    String destCardInv = tableName("cardB_Inv");

    createTable(srcCardInv, TBL_SP_CARD);
    computeCard(src, o, p, srcCardInv);

    createTable(destCardInv, TBL_SP_CARD);
    computeCard(dest, o, p, destCardInv);

    // create indexes on card
    createCardIndex(destCardInv);
    createCardIndex(srcCardInv);


    String mapPairs = T_MAP_PAIRS;

    createSigmaIndex(mapPairs);

    // MAIN LOOP OF THE ALGORITHM

    String t0 = mapPairs;
    String ti = T_TI;
    createTable(ti, TBL_SIGMA);
    copySim(t0, ti);

    for(int i=0; i < iterations; i++) {

      long startTime = System.currentTimeMillis();
      System.err.println("-------- Iteration: " + i);

      String tiSum = tableName("tiSum");
      createTable(tiSum, TBL_SIGMA);
      
      createSigmaIndex(ti); // needed only because outer join not supported in MySQL!

      addSim(tiSum, ti, t0); // tiSum = ti+t0

      if(!PRESERVE_TABLES_FOR_DEBUG)
	dropTable(ti);

      // create index on sum
      if(!STRAIGHT_FLOW_JOIN)
	createSigmaIndex(tiSum);
      
      // tiPhi = phi(ti+t0)

      String tiPhiDupl = tableName("tiPhiDupl");
      createTable(tiPhiDupl, TBL_SIGMA);
      flow(tiPhiDupl, tiSum, src, dest, s, p, o, srcCardInv, destCardInv);
      flowInv(tiPhiDupl, tiSum, src, dest, o, p, s, srcCard, destCard);
      
      String tiPhi = tableName("tiPhi");
      createTable(tiPhi, TBL_SIGMA);

      execute("INSERT INTO " + tiPhi +
	      " SELECT node_a,node_b,sum(sim) FROM " + tiPhiDupl + " GROUP BY node_a,node_b");

      if(!PRESERVE_TABLES_FOR_DEBUG)
	dropTable(tiPhiDupl);

      createTable(ti, TBL_SIGMA);

      createSigmaIndex(tiPhi); // both needed only because outer join not supported in MySQL!
      createSigmaIndex(tiSum);

      addSim(ti, tiPhi, tiSum);

      if(!PRESERVE_TABLES_FOR_DEBUG) {
	dropTable(tiSum);
	dropTable(tiPhi);
      }
      
      normalize(ti, "sim");

      System.err.println("======== Iteration: " + i + " took " +
			 (System.currentTimeMillis() - startTime) / 60000 + " min");

    }

  }

  public static void main(String[] args) throws Exception {

    if(args.length != 3) {
      System.err.println("Usage: SimFloodSQL <table 1> <table 2> <max it>");
      System.exit(1);
    }

    String src = args[0];
    String dest = args[1];
    int iterations = Integer.parseInt(args[2]);

    SimFloodSQL sfa = new SimFloodSQL();

    sfa.defaultMapPairs(src, dest);
//     sfa.createSPOIndex(src);
//     sfa.createSPOIndex(dest);

    sfa.mjoin(src, dest, iterations);
  }

}
