package com.interdataworking.mm.alg;

import com.interdataworking.*;
import com.interdataworking.mm.*;
import com.interdataworking.mm.alg.*;
import com.interdataworking.mm.interpreter.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import java.util.*;

/**
 * Given a set of forward and backward arcs, generates a map of
 * <obj, dependent obj>
 **/

public class Span implements UntypedGateway {

  public List execute(List input) throws ModelException {

    Model src = (Model)input.get(0);
    Model selectorM = (Model)input.get(1);
    Set selector = MapVocabulary.asSet(selectorM);
    Model allM = (Model)input.get(2);
    Set all = MapVocabulary.asSet(allM);

    Model result = src.create();

    Set marked = new HashSet(selector);

    for(Iterator it = selector.iterator(); it.hasNext();) {
      Resource n = (Resource)it.next();
      System.err.println("Starting from " + n);
      collectProp(n, src, marked, new HashSet(), all);
    }

    ArrayList l = new ArrayList();
    l.add(MapVocabulary.asModel(selectorM.create(), marked));
    return l;
  }

  static final void collectSpan(Resource n, Model src, Set marked, Set nodeStack, Set all) throws ModelException {

    if(/*!all.contains(n) ||*/ !nodeStack.add(n))
      return; // make sure no node is visited twice on the same path

    System.err.println("Visiting span: " + n);

    for(Enumeration en = src.find(n, null, null).elements(); en.hasMoreElements();) {

      Statement st = (Statement)en.nextElement();
      if(st.object() instanceof Literal)
	continue;

      if(marked.contains(st.object()) && !nodeStack.contains(st.object())) {

	System.err.println("Reached " + st.object() + ", adding: " + nodeStack);
	marked.addAll(nodeStack);

      } else { // blank
	collectSpan((Resource)st.object(), src, marked, nodeStack, all);
      }
    }

    for(Enumeration en = src.find(null, null, n).elements(); en.hasMoreElements();) {

      Statement st = (Statement)en.nextElement();

      if(marked.contains(st.subject()) && !nodeStack.contains(st.subject())) {

	System.err.println("Reached " + st.subject() + ", adding: " + nodeStack);
	marked.addAll(nodeStack);

      } else { // blank
	collectSpan(st.subject(), src, marked, nodeStack, all);
      }
    }

    nodeStack.remove(n);
  }


  static final void collectProp(Resource n, Model src, Set marked, Set nodeStack, Set all) throws ModelException {

    if(/*!all.contains(n) ||*/ !nodeStack.add(n))
      return; // make sure no node is visited twice on the same path

    System.err.println("Visiting prop: " + n);

    for(Enumeration en = src.find(n, null, null).elements(); en.hasMoreElements();) {

      Statement st = (Statement)en.nextElement();
      if(st.object() instanceof Literal)
	continue;

      if(marked.contains(st.object()) && !nodeStack.contains(st.object())) {

	System.err.println("Reached prop " + st.object() + ", adding: " + nodeStack);
	marked.addAll(nodeStack);

      } else { // blank
	collectProp((Resource)st.object(), src, marked, nodeStack, all);
      }
    }

    nodeStack.remove(n);
  }
  

  public int getMinInputLen() { return 3; }

  public int getMaxInputLen() { return 3; }

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }
  
}
