package com.interdataworking.mm.gui;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.border.*;
//import javax.swing.colorchooser.*;
//import javax.swing.filechooser.*;
import javax.swing.table.*;
import javax.accessibility.*;
import java.text.DecimalFormat;

import java.awt.Graphics2D;
import java.awt.Graphics;
import java.awt.RenderingHints;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.geom.*;
// import java.awt.event.*;
// import java.beans.*;
import java.util.*;
import java.io.*;
// import java.applet.*;
import java.net.*;
import com.interdataworking.mm.*;
import com.interdataworking.mm.alg.*;

public class ArrowPanel extends JPanel {

  boolean WEIGHTS_IN_PERCENT = true;
  boolean WEIGHTS_ON_RED_ARCS = false;
  boolean WEIGHTS_ON_GRAY_ARCS = true;

  static final DecimalFormat nf = new DecimalFormat("##.00");

  ChangeListener listener;
  JScrollPane lscroll, rscroll;
  Map l2j, r2j;
  Map map, multimap;
  boolean drawArrows = false; // flag

  public ArrowPanel() {

    super();

    listener = new ChangeListener() {

	public void stateChanged(ChangeEvent e) {

	  //	  new Exception().printStackTrace(System.err);
	  repaint();
	}
      };
  }

  public ChangeListener getRepaintListener() {

    return listener;
  }

  public void setScrollPanes(JScrollPane left, JScrollPane right) {

    this.lscroll = left;
    this.rscroll = right;
    lscroll.getViewport().addChangeListener(listener);
    rscroll.getViewport().addChangeListener(listener);
  }

  public void setMappings(Map map, Map multimap, Map l2j, Map r2j) {

    //    System.err.println("Set mappings: red=" + map.size() + ", gray=" + multimap.size());
    this.map = map;
    this.multimap = multimap;
    this.l2j = l2j;
    this.r2j = r2j;
  }

  public void setDrawArrows(boolean b) {

    drawArrows = b;
  }

  public static void move(Rectangle2D r, double offX, double offY) {

    r.setRect(r.getX() + offX, r.getY() + offY,
	      r.getWidth(), r.getHeight());
  }

  static void subtractBoxes(Area clipArea, Collection leftBoxes, double deltaX, double deltaY) {

    //    System.err.println("Clip area: " + clipArea);
    for(Iterator it = leftBoxes.iterator(); it.hasNext();) {
      
      JComponent c = (JComponent)it.next();
      Rectangle2D cbounds = c.getBounds();
      move(cbounds, deltaX, deltaY);
      System.err.println("Subtracting " + cbounds);
      clipArea.subtract(new Area(cbounds));
    }
  }

  public void paint(Graphics g) {
    
    synchronized (this) {
      super.paint(g);
      Graphics2D g2d = (Graphics2D) g;

      Rectangle2D lbounds = lscroll.getViewport().getBounds();
      Rectangle2D rbounds = rscroll.getViewport().getBounds();

      move(lbounds, lscroll.getX(), lscroll.getY());
      move(rbounds, rscroll.getX(), rscroll.getY());

      Point2D loffset = (Point2D)lscroll.getViewport().getViewPosition();
      Point2D roffset = (Point2D)rscroll.getViewport().getViewPosition();

      double ldeltaX = -loffset.getX() + lscroll.getX();
      double ldeltaY = -loffset.getY() + lscroll.getY();
      double rdeltaX = -roffset.getX() + rscroll.getX();
      double rdeltaY = -roffset.getY() + rscroll.getY();

      //      Area clipArea = new Area(lbounds);
      //      clipArea.add(new Area(rbounds));

      //      Area clipArea = new Area(lbounds.createUnion(rbounds));

      //      System.err.println("Left area: " + lbounds);
      //      Area leftArea = new Area(lbounds);
      // subtract all boxes
      //      subtractBoxes(leftArea, leftBoxes, ldeltaX, ldeltaY);

      //      Area rightArea = new Area(rbounds);
      //      subtractBoxes(rightArea, rightBoxes, rdeltaX, rdeltaY);

      // leftArea.add(rightArea);
      // g2d.setClip(leftArea);

      g2d.setClip(lbounds.createUnion(rbounds));
      
      g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			   RenderingHints.VALUE_ANTIALIAS_ON);

      Color grayedColor = new Color(100, 100, 100);
      Color selectedColor = Color.red;
      //      g2d.setColor(grayedColor);

      //      BasicStroke solid = new BasicStroke(9.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND, 9.0f);
      //      g2d.setStroke(solid);

      //      g2d.setComposite(opaque);
      
      g2d.setColor(grayedColor);
      drawMap(multimap, g2d, ldeltaX, ldeltaY, rdeltaX, rdeltaY, WEIGHTS_ON_GRAY_ARCS, WEIGHTS_IN_PERCENT);

      g2d.setColor(selectedColor);
      drawMap(map, g2d, ldeltaX, ldeltaY, rdeltaX, rdeltaY, WEIGHTS_ON_RED_ARCS, false);
    }
// 	if (img != null) {
// 	  int imgw = img.getWidth();
// 	  int imgh = img.getHeight();
// 	  g2d.setComposite(AlphaComposite.Src);
// 	  g2d.drawImage(img, null, 0, 0);
// 	}
//       }
  }

  void drawMap(Map map, Graphics2D g2d, double ldeltaX, double ldeltaY, double rdeltaX, double rdeltaY, boolean weights, boolean weightsInPercent) {

    // adjust coordinates
    if(l2j == null || r2j == null)
      return; // not yet initialized

    for(Iterator it = map.values().iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();

      JComponent lc = (JComponent)l2j.get(p.getLeft());
      JComponent rc = (JComponent)r2j.get(p.getRight());

      if(lc == null || rc == null)
	continue;

      double x1 = lc.getX() + lc.getWidth()/2 + ldeltaX;
      double y1 = lc.getY() + lc.getHeight()/2 + ldeltaY;

      double x2 = rc.getX() + rc.getWidth()/2 + rdeltaX;
      double y2 = rc.getY() + rc.getHeight()/2 + rdeltaY;

      // make sure the ends lie in the different panes

      //	System.err.println("lbounds = " + lbounds + ", rbounds = " + rbounds);
      //	System.err.println("lbounds2 = " + lscroll.getBounds() + ", rbounds2 = " + rscroll.getBounds());

      //	boolean bothInLeft = lbounds.contains(x1, y1) && lbounds.contains(x2, y2);
      //	boolean bothInRight = rbounds.contains(x1, y1) && rbounds.contains(x2, y2);

      //	if(!bothInLeft && !bothInRight) {

      if(drawArrows) {
	if(!p.inverse)
	  GUIUtil.drawArrow(g2d, (int)x1, (int)y1, (int)x2, (int)y2, 7, 7, null, null, false);
	else
	  GUIUtil.drawArrow(g2d, (int)x2, (int)y2, (int)x1, (int)y1, 7, 7, null, null, false);
      }

      g2d.drawLine((int)x1, (int)y1, (int)x2, (int)y2);

      //      Line2D l = new Line2D.Double(x1, y1, x2, y2);
      //	  System.err.println("line = (" + x1 + "," + y1 + ")-(" + x2 + "," + y2 + ")");
      //      g2d.draw(l);

      if(weights) { // place weight indicator

	String ws = weightsInPercent ? String.valueOf((int)Math.round(p.sim * 100)) + "%" : nf.format(p.sim);
	g2d.drawString(ws, (float)(x1+x2)/2, (float)(y1+y2)/2);
      }
      //	}
    }
  }
}
