package com.interdataworking.util;

import java.util.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;
import org.w3c.rdf.util.*;
import org.xml.sax.SAXException;

import com.interdataworking.*;
import com.interdataworking.mm.*;

public class IDWUtil {

  static void bailOut() {

    System.err.println("Usage: IDWUtil <file name> <parser class name> {<gateway class name>}+");
    System.exit(1);
  }

  public static List toParam(Model param) {

    List l = new ArrayList();
    l.add(param);
    return l;
  }

  public static Model toModel(List l) {

    return (Model)l.get(0);
  }

  public static void flatten(List l, List result) {

    for(int i=0; i < l.size(); i++) {

      Object n = l.get(i);
      if(n instanceof List)
	flatten((List)n, result);
      else
	result.add(n);
    }
  }

  public static List apply(UntypedGateway g, List l) throws ModelException {

    List result = new ArrayList();
    flatten(l, result);
    return g.execute(result);
  }

  public static List apply(UntypedGateway g, List l1, List l2)  throws ModelException {

    List ofLists = new ArrayList();
    ofLists.add(l1);
    ofLists.add(l2);

    return apply(g, ofLists);
  }

  public static Model apply1(UntypedGateway g, Model m, Set selector, Model map)  throws ModelException {

    return toModel(apply(g,
			 IDWUtil.toParam(m),
			 IDWUtil.toParam(MapVocabulary.asModel(m.create(), selector)),
			 IDWUtil.toParam(map)));
  }

  public static List apply(UntypedGateway g, List l1, List l2, List l3)  throws ModelException {

    List ofLists = new ArrayList();
    ofLists.add(l1);
    ofLists.add(l2);
    ofLists.add(l3);

    return apply(g, ofLists);
  }

  public static List apply(UntypedGateway g, Model m)  throws ModelException {

    return g.execute(toParam(m));
  }

  public static void main(String[] args) throws Exception {

    if(args.length < 3)
      bailOut();

    String filename = args[0];
    RDFFactory f = new RDFFactoryImpl();
    Model m = f.createModel();

    RDFParser parser = (RDFParser)Class.forName(args[1]).newInstance();

    try {
      RDFUtil.parse(filename, parser, m);
    } catch (SAXException sex) {
      sex.getException().printStackTrace(System.err);
      System.exit(1);
    }

    for(int i = 2; i < args.length-1; i++) {

      ModelGateway gw = (ModelGateway)Class.forName(args[i]).newInstance();
      System.err.println("Applying " + gw);
      gw.init(null, null);
      m = gw.translate(m);
    }

    // determine serializer
    RDFSerializer sr = (RDFSerializer)Class.forName(args[args.length-1]).newInstance();

    //    RDFUtil.printStatements(m, System.out);
    RDFUtil.dumpModel(m, System.out, sr);

    System.err.println("Done.");
  }
}
