package edu.stanford.db.alg.ngram;

import java.util.*;

/**
 * Used for
 * a) chopping n-grams (then, element is an n-gram string),
 * b) for "document" entries (then, element is a user object)
 * c) and for NGram entries (that hold overall ngram frequencies)
 */

public class Entry {

  Object element;
  int occ;

  public Entry(Object element, int occ) {

    this.element = element;
    this.occ = occ;
  }

  public Entry(Object element) {

    this(element, 0);
  }

  public Object getElement() {

    return element;
  }

  public void setElement(Object el) {

    this.element = el;
  }

  public void incCount() {

    occ++;
  }

  public void incCount(int value) {

    occ += value;
  }

  public int getCount() {

    return occ;
  }

  public int hashCode() {

    return element.hashCode();
  }

  public boolean equals(Object that) {

    if(this == that)
      return true;

    if(!(that instanceof Entry))
      return false;

    return element.equals(((Entry)that).element);
  }

  public String toString() {

    return "[E: " + element + ", " + occ + "]";
  }

  public static Comparator getCountComparator() {

    return (new Entry(null)).new CountComparator();
  }

  public static Comparator getElementComparator() {

    return (new Entry(null)).new ElementComparator();
  }

  public static Comparator getLengthComparator() {

    return (new Entry(null)).new LengthComparator();
  }

  public class CountComparator implements Comparator {

    public int compare(Object o1, Object o2) {

      Entry e1 = (Entry)o1;
      Entry e2 = (Entry)o2;
      
      if(e1.occ < e2.occ)
	return -1;
      else if(e1.occ > e2.occ)
	return 1;
      
      // else: element comparison
      
      return ((Comparable)e1.element).compareTo(e2.element);
    }

    public boolean equals(Object that) {

      return this == that;
    }
  }

  public class ElementComparator implements Comparator {

    public int compare(Object o1, Object o2) {

      Entry e1 = (Entry)o1;
      Entry e2 = (Entry)o2;
      
      return ((Comparable)e1.element.toString()).compareTo(e2.element.toString());
    }

    public boolean equals(Object that) {

      return this == that;
    }
  }

  public class LengthComparator implements Comparator {

    public int compare(Object o1, Object o2) {

      String e1 = ((Entry)o1).element.toString();
      String e2 = ((Entry)o2).element.toString();

      if(e1.length() < e2.length())
	return 1;
      else if(e1.length() > e2.length())
	return -1;
      
      // else: element comparison
      
      return e1.compareTo(e2);
    }

    public boolean equals(Object that) {

      return this == that;
    }
  }
}
