/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package edu.stanford.db.rdf.query.memory;

import org.w3c.rdf.model.*;
import edu.stanford.db.rdf.query.*;
import org.w3c.rdf.util.RDFUtil;

import java.util.*;

/**
 * This is a query iterator.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu> */

public class NodeResultSet extends Hashtable implements NodeIterator {
    
    QueryIterator it;
    // in case if variables have been set
    boolean needsReset = true;
    // only needed to pass further down the foodchain
    Model model;

    public NodeResultSet(QueryIterator it) throws ModelException {
	
	this.it = it.duplicate();
    }

    public NodeResultSet(QueryIterator it, Model m) throws ModelException {
	
	this(it);
	model = m;
    }

    public NodeResultSet(QueryIterator[] plan) throws ModelException {
	
	this(new JoinIterator(plan));
    }

    public NodeResultSet(QueryIterator[] plan, Model m) throws ModelException {

	this(plan);
	model = m;
    }

    public int getCost() {

	return it.getCost();
    }

    public void reset() throws ModelException {

	clear();
	needsReset = true;
    }

    public void reset(Model m) throws ModelException {

	clear();
	model = m;
	needsReset = true;
    }

    public void set(String varName, RDFNode value) throws ModelException {

	super.put(varName, value);
	needsReset = true;
    }

    public boolean next() throws ModelException {
	
	if(needsReset) {
	    it.reset(this, model);
	    needsReset = false;
	}
	return it.next(this);
    }

    public RDFNode get(String varName) {

	return (RDFNode)super.get(varName);
    }

  public Map getAll() {

    return this;
  }
}
